__author__ = 'matthew'

import numpy as np
from matplotlib import pyplot as plt
import argparse
from scipy.optimize import curve_fit
from scipy.stats import linregress
import math

parser = argparse.ArgumentParser(description="Viscosity time decomposition analysis")
parser.add_argument('files', nargs='+', help="list all visco*.xvg files")
parser.add_argument('-c', default=10000, type=int, help="chunk size for determining zero slope plateau time, default 10000 energy frames")
parser.add_argument('-s', default=0.01, type=float, help="slope tolerance for determining zero slope plateau time, default 0.01")
parser.add_argument('-i', default=2, type=float, help="ignore this beginning amount of time (in ps) for fit to running integral, default 2 ps")
parser.add_argument('-g', action='store_true', help="display graphs. Default is no.")
args = parser.parse_args()

files = args.files
chunksize = args.c
slope_tolerance = args.s
begin_fit_time = args.i
graphs = args.g

nblocks = len(files)


# loading in running integral data for each file
# note that times are same in each file, so take from 1st file (i.e. time is 1D array)
# y-values are taken from all files, so y is a 2D array with nblocks columns)
time = np.loadtxt(fname=files[0], skiprows=25)[:,0]
print(time[0])
dt = time[1] - time[0]
y = np.empty(shape=[len(time), nblocks])
for j in range(0, nblocks):
	y[:,j] = np.loadtxt(fname=files[j], skiprows=25)[:len(time),1]

# find average running integral and stdev as function of time from other running integrals
# axis 1 means analyze across a row over the columns
print()
print("Determining mean and standard deviation (sigma) of running integrals as functions of time.")
print()
runningint = np.mean(y, axis=1) 
sigma = np.std(y, axis=1, ddof=1)

# fitting sigma(t) to y = a*t^b
def power_law_fit(t, a, b):
	return a * t ** b

sigma_fit_popt, sigma_fit_pcov = curve_fit(power_law_fit, xdata=time, ydata=sigma)
a = sigma_fit_popt[0]
b = sigma_fit_popt[1]
sigma_fit = power_law_fit(time, *sigma_fit_popt)
print("Fitting sigma to power law function.")
print(f"Power law fit: y = {round(a, 3)}*t^{round(b, 3)}")
print()

# need to find zero slope for plateau time and eventually cutoff time for fitting
# define function that returns the index with 0 slope within some slope tolerances
# purpose is to find flat region of running integral
def zero_slope_index(x, y, chunksize, max_slope):
	for index in range(0, len(x) - chunksize + 1):
		result = linregress(x[index:(index + chunksize)], y[index:(index + chunksize)]) 
		slope = result.slope
		if 0 <= abs(slope) <= max_slope:
			return int((index + (index + chunksize)) / 2)
	# if criteria is not met, function returns nothing
	return None;

# run through a bunch of different max_slope in increments of 0.001 up until slope_tolerance
# if a slope of approximately zero is found, the time which that occurs will be reported
# if no slope is found within the specified slope_tolerance on command line, stop script. 
max_slope = 0
while max_slope <= slope_tolerance:
	flat_start_index = zero_slope_index(time, runningint, chunksize=chunksize, max_slope=max_slope)
	if flat_start_index != None:
		print(f"Zero slope achieved at {time[flat_start_index]} ps using chunks of {chunksize} energy frames "
			f"({chunksize*dt} ps) with max slope {max_slope} (mPa*s)/s.")
		break
	max_slope += 0.001

if max_slope > slope_tolerance:
	print(f"Cannot determine plateau time based on current slope tolerance of {slope_tolerance} (mPa*s)/s.")
	print("Either increase the number of blocks, increase the slope tolerance, or decrease chunk size. The first option is preferable.")
	print("Stopping script.")
	exit()

# find mean of flat running integral region and use that for time cutoff determination
runningint_flat_avg = np.mean(runningint[flat_start_index:])
print(f"Average of running integral after zero slope time {time[flat_start_index]} ps (flat region) is {round(runningint_flat_avg, 4)} mPa*s.")

# t_cutoff is chosen such that sigma(t_cutoff) = 0.4*mean(running integral flat region)
print(f"t_cutoff should be chosen such that sigma(t_cutoff) is about 40% of the average over the flat region of {round(runningint_flat_avg, 4)} mPa*s.")
cutoff_sigma = 0.4*runningint_flat_avg
cutoff_time_index_actual=0
print("Calculating cutoff time with exactly 40% criterion.")
while(True):
	if sigma[cutoff_time_index_actual] > cutoff_sigma:
		break
	cutoff_time_index_actual += 1
print(cutoff_time_index_actual)
print(f"cuttoff_time_actual = {time[cutoff_time_index_actual]} ps")
cutoff_time_actual = time[cutoff_time_index_actual] # actual cutoff time
print(cutoff_time_actual)
print(np.ceil(cutoff_time_actual / 100))
cutoff_time_rounded = np.ceil(cutoff_time_actual / 100) * 100 # cutoff time rounded upwards to the hundred; use this 
print(cutoff_time_rounded)
cutoff_time_index_rounded = np.where(abs(time - cutoff_time_rounded)< 0.005)[0][0]
print(cutoff_time_index_rounded)

print(f"Actual cutoff time is {cutoff_time_actual} ps. Will round upwards to nearest hundred to "
	f"{cutoff_time_rounded} ps and use this as the upper limit for the double exp fit to running integral.")
print(f"sigma({cutoff_time_rounded} ps) = {round(sigma[cutoff_time_index_rounded], 3)} mPa*s is only "
	f"{np.round(sigma[cutoff_time_index_rounded]/runningint_flat_avg*100, 1)}% of the average over the flat region of running integral of {round(runningint_flat_avg, 4)} mPa*s.")
print("If the percentage is between 30-50%, result is likely satisfactory.")	
print(f"Please check that the rounded cutoff time of {cutoff_time_rounded} ps is reasonable and allows the double exp fit to the running integral to cover a moderately reasonable amount of the plateau region.")
print(f"Also please check that the size (time length) of your blocks are at least 10x that of the plateau time {time[flat_start_index]} ps. If not, then the running integral may not converge properly.")
print("However, if increasing the number of blocks is more important (for good statistics) than the length itself, it may be relevant to disregard the previous check within reason.")
print()


# Fitting running integral to complex double exponential function with following conditions 
# 1. ignore first few picoseconds (let's do 2 ps as default) 
# 2. weight according to 1/t^b 
# note that weight = 1 / "sigma"^2 in curve_fit ("sigma" is curve_fit sigma), so "sigma" = 1/sqrt(weight) = t^(b/2)
def double_exp_fit(t, A, alpha, tau_1, tau_2):
	return A*alpha*tau_1*(1 - np.exp(-t/tau_1)) + A*(1 - alpha)*tau_2*(1 - np.exp(-t/tau_2))

weight = 1 / np.power(time, b)
begin_fit_index = np.where(abs(time - begin_fit_time)< 0.005)[0][0]
end_fit_index = np.where(abs(time - cutoff_time_rounded)< 0.005)[0][0]

print(f"Will fit the running integral to a double exp function from {time[begin_fit_index]} ps to {time[end_fit_index]} ps with weight 1/t^{round(b, 3)}.")
	
runningint_fit_popt, runningint_fit_pcov = curve_fit(double_exp_fit, xdata=time[begin_fit_index:(end_fit_index + 1)],
	ydata=runningint[begin_fit_index:(end_fit_index + 1)], sigma = 1/np.sqrt(weight[begin_fit_index:(end_fit_index + 1)]), absolute_sigma = False)
A = runningint_fit_popt[0]
alpha = runningint_fit_popt[1]
tau_1 = runningint_fit_popt[2]
tau_2 = runningint_fit_popt[3]
runningint_fit = double_exp_fit(time, *runningint_fit_popt)
print(f"Double exp fit: y = {round(A, 3)}*{round(alpha, 3)}*{round(tau_1, 3)}*(1 - e^(-t/{round(tau_1, 3)})) + "
	f"{round(A, 3)}*(1 - {round(alpha, 3)})*{round(tau_2, 3)}*(1 - e^(-t/{round(tau_2, 3)}))")

viscosity_limit = A*alpha*tau_1 + A*(1-alpha)*tau_2
print("Viscosity is calculated by taking the infinite time limit of the double exp fit.")
print(f"Viscosity: {viscosity_limit} mPa*s")
print()


fig, (ax1, ax2) = plt.subplots(1,2, figsize=(8, 4))
ax1.set_xlabel("Autocorrelation time (ps)")
ax1.set_ylabel("$\eta$ (mPa*s)")
ax1.plot(time, runningint, color='black', linestyle='solid', label="Mean running integral")
ax1.plot(time, runningint_fit, color='red', linestyle='solid', label="Fitted double exp function")
ax1.legend(loc='upper left', prop={'size':6})

left, bottom, width, height = [0.25, 0.2, 0.2, 0.2]
ax3 = fig.add_axes([left, bottom, width, height])
ax3.set_xlabel("Autocorrelation time (ps)")
ax3.set_ylabel("$\eta$ (mPa*s)")
five_ps_index = np.where(abs(time - 5)< 0.005)[0][0]
ax3.plot(time[0:(five_ps_index + 1)], runningint[0:(five_ps_index + 1)], color='black', linestyle='solid', label="Mean running integral")


ax2.set_xlabel("Autocorrelation time (ps)")
ax2.set_ylabel("$\sigma$ (mPa*s)")
ax2.plot(time, sigma, color='black', linestyle='solid', label="Standard deviation")
ax2.plot(time, sigma_fit, color='red', linestyle='solid', label="Power law fit a*t^b")
ax2.legend(loc="upper left", prop={'size':6})

if graphs == True:
	plt.show()
